#!/bin/sh
#
# Test for 3 different latencies, for traditional copy and copy offload.
# Test for different file sizes.
#
# Copyright 2010 NetApp, Inc.,
# contribution by Anshul Madan
# contribution by James Lentini
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License v2 for more details.

CAT=/bin/cat
CP=/bin/cp
CUT=/bin/cut
DATE=/bin/date
DT=dt
ECHO=/bin/echo
ETHTOOL=/sbin/ethtool
GREP=/bin/grep
HELPER=./copy-perf-helper.sh
IP=/sbin/ip
MKTEMP=/bin/mktemp
MOUNT=/bin/mount
NETSTAT=/bin/netstat
PING=/bin/ping
RM=/bin/rm
SED=/bin/sed
SSH=/usr/bin/ssh
TIME=/usr/bin/time
TR=/usr/bin/tr
VMSTAT=/usr/bin/vmstat

#
# Function: copyFile
#   Params: <src>, source file name
#           <dst>, destination file name
#           <server>, DNS name of NFS server
#           <type>, local, traditional, or offload
#           <size, source file size
#           <speed>, network speed
#           <log>, file name for statistics
#
copyFile() {
	if [ 7 -ne $# ]; then
		$ECHO "incorrect arguments to $FUNCNAME"
		exit 1
	fi

	local src=$1
	local dst=$2
	local server=$3
	local type=$4
	local size=$5
	local speed=$6
	local log=$7

	local flags=""
# Uncomment to read after copy
#	flags="-r"
# Uncomment to clear then read after copy
#	flags="-c -r"

	$ECHO -n "$size $speed $type " >> $log
	if [ "$type" = "local" ]; then
		$SSH "root@$server" "$TIME -f '%e' $CP $src $dst" 2>> $log
	else
		$TIME -f '%e' $HELPER $flags $type $src $dst 2>> $log
	fi
}

#
# Function: recordPacketsStart
#   Params: <log>, the log file name
#           <type>
#           <size>
#           <speed>
#
recordPacketsStart() {
	if [ 4 -ne $# ]; then
		$ECHO "incorrect arguments to $FUNCNAME"
		exit 1
	fi

	local log=$1
	local type=$2
	local size=$3
	local speed=$4
	
	$ECHO -n "$size $speed $type " >> $log
}

#
# Function: recordPacketsStop
#   Params: <log>, the log file name
#
recordPacketsStop() {
	if [ 1 -ne $# ]; then
		$ECHO "incorrect arguments to $FUNCNAME"
		exit 1
	fi

	local log=$1
	
	# End the record by writing a newline
	$ECHO "" >> $log
}

#
# Function: recordPackets
#   Params: <server>
#           <packet_log>
#
# Note: there are a number of ways to access the packet statistics
#
# (1) /sys/class/net/ethX/statistics/rx_bytes
#     where ethX is the Ethernet interface being used (e.g. eth0).
#     The bytes are also available: /sys/class/net/ethX/statistics/tx_bytes
#
# (2) netstat(1):
#     $ netstat -e -e --interfaces=ethX
#
# (3) ifconfig(1):
#     $IFCONFIG eth1 | $GREP 'RX packets' | $SED 's/^[ \t]*//' | $CUT -d ':' -f 2 | $CUT -d ' ' -f 1
#
recordPacketsMeasure() {
	if [ 3 -ne $# ]; then
		$ECHO "incorrect arguments to $FUNCNAME"
		exit 1
	fi

	local server=$1
	local log=$2
	local iface=$3

	local rx_cmd="$CAT /sys/class/net/$iface/statistics/rx_bytes"
	local tx_cmd="$CAT /sys/class/net/$iface/statistics/tx_bytes"

	# Measure the client's packet counts
	local rx_packets=`eval ${rx_cmd}`
	local tx_packets=`eval ${tx_cmd}`
	$ECHO -n "$rx_packets $tx_packets " >> $log

	# Measure the server's packet counts
	$SSH "root@$server" "${rx_cmd}" | $TR -d '\n' >> $log
	$ECHO -n " " >> $log
	$SSH "root@$server" "${tx_cmd}" | $TR -d '\n' >> $log
	$ECHO -n " " >> $log
}

#
# Function: clearCaches
#   Params: <server>
clearCaches() {
	if [ 1 -ne $# ]; then
		$ECHO "incorrect arguments to $FUNCNAME"
		exit 1
	fi

	local server=$1

	$ECHO 3 > /proc/sys/vm/drop_caches
	if [ 0 -ne $? ]; then
		$ECHO "Failed to clear cache on NFS client"
		exit 1
	fi

	$SSH "root@$server" "$ECHO 3 > /proc/sys/vm/drop_caches"
	if [ 0 -ne $? ]; then
		$ECHO "Failed to clear cache on NFS server $server"
		exit 1
	fi
}

#
# Function: runTest
#   Params: <mount_dir>
#           <local_dir>
#           <server>
#           <iface>
#           <time_log>
#           <packet_log>
#           <type>
#           <size>
#           <speed>
runTest() {
	if [ 9 -ne $# ]; then
		$ECHO "incorrect arguments to $FUNCNAME"
		exit 1
	fi

	local mount_dir=$1
	local local_dir=$2
	local server=$3
	local iface=$4
	local time_log=$5
	local packet_log=$6
	local type=$7
	local size=$8
	local speed=$9

	local src="${mount_dir}/src-${size}MB"
	local dst="${mount_dir}/dst"

	# if the source file does not exist or is not a regular file
	if [ ! -f $src ]; then
		$DT of=$src pattern=incr bs=4k limit=${size}m dispose=keep > /dev/null
		if [ 0 -ne $? ]; then
			$ECHO "Failed to create source file $src"
			exit 1
		else
			$ECHO "Created source file $src"
		fi
	else
		$ECHO "Reusing source file $src"
	fi

	$RM -f $dst

	recordPacketsStart $packet_log $type $size $speed
	recordPacketsMeasure $server $packet_log $iface

	if [ "$type" = "local" ]; then
		src="${local_dir}/src-${size}MB"
		dst="${local_dir}/dst"
	fi

	copyFile $src $dst $server $type $size $speed $time_log

	recordPacketsMeasure $server $packet_log $iface
	recordPacketsStop $packet_log

	clearCaches $server
}

#
# Function: getServer
#   Params: <mount_dir>
getServer() {
	if [ 1 -ne $# ]; then
		$ECHO "incorrect arguments to $FUNCNAME"
		exit 1
	fi

	local mount_dir=$1

	$MOUNT -t nfs4 | $GREP "on $mount_dir type" | $CUT -d ' ' -f 1 | $CUT -d ':' -f 1
}

#
# Function: getIface
#   Params: <server>
getIface() {
	if [ 1 -ne $# ]; then
		$ECHO "incorrect arguments to $FUNCNAME"
		exit 1
	fi

	local server=$1
	local ip=`$PING -c 1 -n $server | $GREP "bytes from" | $CUT -d ' ' -f 4 | $SED 's/://g'`

	$IP route get $ip | $GREP dev | $SED 's/.*dev //g' | $CUT -d ' ' -f 1
}

#
# Main
#

if [ 2 -ne $# ]; then
	$ECHO "Usage: $0 <mount_path> <local_path"
	$ECHO "Where"
	$ECHO -e "\t mount_path : the path to an NFS mount point"
	$ECHO -e "\t local_path : the export path on the NFS server"
	exit 1
fi

mount_dir=`$ECHO "$1" | sed 's|/*$||'` # strip any trailing '/' characters
local_dir=`$ECHO "$2" | sed 's|/*$||'` # strip any trailing '/' characters

# get the DNS name of the NFS server
server=$(getServer $mount_dir)

# get the network interface on the client
iface=$(getIface $server) 

log_template="perf_logs_`$DATE +%Y_%m_%d_%k_%M`.XXXXXX"
log_dir=`$MKTEMP -d --tmpdir=$PWD $log_template`
$ECHO "Created log directory $log_dir"

#
# Create the time data log
#
time_log="${log_dir}/time.dat"
$ECHO "#" >> $time_log
$ECHO "# File: $time_log" >> $time_log
$ECHO "#" >> $time_log
$ECHO "# Client mount: $mount_dir" >> $time_log
$ECHO "# Server export: $local_dir" >> $time_log
$ECHO "# Server: $server" >> $time_log
$ECHO "# Interface: $iface" >> $time_log
$ECHO "#" >> $time_log
$ECHO "# Format: <size (MB)> <network speed (mbps)> <local|traditional|offload> <time (sec)>" >> $time_log
$ECHO "#" >> $time_log
$ECHO "" >> $time_log 

#
# Create the packet data log
#
packet_log="${log_dir}/packet.dat"
$ECHO "#" >> $packet_log
$ECHO "# File: $packet_log" >> $packet_log
$ECHO "#" >> $packet_log
$ECHO "# Client mount: $mount_dir" >> $packet_log
$ECHO "# Server export: $local_dir" >> $packet_log
$ECHO "# Server: $server" >> $packet_log
$ECHO "# Interface: $iface" >> $packet_log
$ECHO "#" >> $packet_log
$ECHO "# Format: <size (MB)> <network speed (mbps)> <local|traditional|offload> <client RX before> <client TX before> \\" >> $packet_log
$ECHO -e "#\t <server RX before> <server TX before> <client RX after> <client TX after> <server RX after> <server TX after>" >> $packet_log
$ECHO "#" >> $packet_log
$ECHO "" >> $packet_log 

# NOTE: changing the speed should be done infrequently.
# Keep this as the outermost loop.
# <FIXME>
#for speed in 1000 10
#do
	# Set the speed of the network interface
#	$ETHTOOL -s $iface speed $speed
	speed=1000
#</FIXME>

	for (( size=2; size <= 2048; size*=2 ))
	do
		for (( iterations=0; iterations < 3; iterations++))
		do
			for type in "local" "traditional" "offload"
			do
				runTest $mount_dir $local_dir $server $iface $time_log $packet_log $type $size $speed 
			done
		done
	done

	# Two blank lines create a new data set
	$ECHO "" >> $time_log
	$ECHO "" >> $time_log

	$ECHO "" >> $packet_log
	$ECHO "" >> $packet_log
#FIXME done
