#!/bin/sh
#
# An automated test of the copyfile(2) system call
#
#
# Copyright 2010 NetApp, Inc., 
# contribution by James Lentini
#
# This program is free software; you can redistribute it and/or 
# modify it under the terms of the GNU General Public License 
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied 
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See the GNU General Public License v2 for more details.

CHMOD=chmod
COPY=copy
DATE=date
DIFF=diff
ECHO=echo
EXPR=expr
MKTEMP=mktemp
PWD=pwd
RM=rm
TOUCH=touch

SEPARATOR="----------------------------------------"

#
# Function: testPrintStart
#   Params: <test_count>
#
printTestStart () {
	if [ 1 -ne $# ]; then
		$ECHO "incorrect arguments to $FUNCNAME"
		exit 1
	fi

	local test_count=$1
	$ECHO -n " $test_count: "
}

#
# Function: testPrintPass
#
printTestPass () {
	$ECHO "Passed"
}

#
# Function: expectSuccess
#   Params: <command> <exit status>
#
expectSuccess () {
	if [ 2 -ne $# ]; then
		$ECHO "incorrect arguments to $FUNCNAME"
		exit 1
	fi

	local cmd=$1
	local rc=$2

	if [ 0 -ne $rc ]; then
		$ECHO "Failure: expected success"
		$ECHO "      command: [$cmd]"
		$ECHO "  exit status: [$rc]"

		return 1
	else
		return 0
	fi
}

#
# Function: expectFailure
#   Params: <command> <exit status>
#
expectFailure () {
	if [ 2 -ne $# ]; then
		$ECHO "incorrect arguments to $FUNCNAME"
		exit 1
	fi

	local cmd=$1
	local rc=$2

	if [ 0 -eq $rc ]; then
		$ECHO "Failure: expected failure"
		$ECHO "      command: [$cmd]"
		$ECHO "  exit status: [$rc]"

		return 1
	else
		return 0
	fi

}

#
# Function: runTest
#   Params: <log file> <expect success> <flags> <source path> <destination path>
#
runTest () {
	if [ 5 -ne $# ]; then
		$ECHO "incorrect arguments to $FUNCNAME"
		exit 1
	fi

	local log=$1
	local expect_success=$2
	local flags=$3
	local src_path=$4
	local dst_path=$5

	local cmd="$COPY -f $flags $src_path $dst_path"
	$cmd >& $log
	local rc=$?

	if [ 1 -eq $expect_success ]; then
		expectSuccess "$cmd" $rc
	else
		expectFailure "$cmd" $rc
	fi

	return $rc
}

#
# Function: verifyCopy
#   Params: <log file> <expect success> <source path> <destination path>
#
verifyCopy() {
	if [ 4 -ne $# ]; then
		$ECHO "incorrect arguments to $FUNCNAME"
		exit 1
	fi

	local log=$1
	local expect_success=$2
	local src_path=$3
	local dst_path=$4

	local cmd="$DIFF $src_path $dst_path"
	$cmd >& $log
	local rc=$?

	if [ 1 -eq $expect_success ]; then
		expectSuccess "$cmd" $rc
	else
		expectFailure "$cmd" $rc
	fi

	return $rc
}

#
# MAIN
#

if [ 2 -ne $# ]; then
	$ECHO "Usage: $0 <source file> <destination directory>"
	exit 1
fi

src_path=$1
dst_dir=$2

# if the file is not a regular file
if [ ! -f $src_path ]; then
	$ECHO "Source file [$src_path] is not a regular file"
	exit 1
fi

#if the file is not readable
if [ ! -r $src_path ]; then
	$ECHO "Source file [$src_path] is not readable"
	exit 1
fi

dst_path=`$MKTEMP --tmpdir=$dst_dir --dry-run copy_dst.XXXXXX`

log_dir=`$PWD`
log_template="log_`$DATE +%Y_%m_%d_%k_%M`.XXXXXX"
log=`$MKTEMP --tmpdir=$log_dir $log_template`

test_count=0
pass_count=0

$ECHO "Test"
$ECHO $SEPARATOR


#
# Test 1
# Full file copy from the source file 
# to a destination file that does not exist
# Expect to succeed
#
test_count=`$EXPR $test_count + 1`
printTestStart $test_count

runTest $log 1 0 $src_path $dst_path
# if the copy succeeded
if [ 0 -eq $? ]; then
	verifyCopy $log 1 $src_path $dst_path
	# if the verify succeeded
	if [ 0 -eq $? ]; then
		printTestPass
		pass_count=`$EXPR $pass_count + 1`
	fi
fi
$RM -f $dst_path


#
# Test 2
# Full file copy from the source file 
# to a destination file that does exist
# Expect to fail
#
test_count=`$EXPR $test_count + 1`
printTestStart $test_count

cmd="$TOUCH $dst_path"
$cmd
expectSuccess "$cmd" $?
# if touch succeeded
if [ 0 -eq $? ]; then
	runTest $log 0 0 $src_path $dst_path
	# if the copy failed
	if [ 0 -ne $? ]; then
		printTestPass
		pass_count=`$EXPR $pass_count + 1`
	fi
fi
$RM -f $dst_path


#
# Test 3
# Full file copy from a source file that does not exist
# to a destination file that does not exist
# Expect to fail
#
test_count=`$EXPR $test_count + 1`
printTestStart $test_count

null_src_path=`$MKTEMP --tmpdir=/tmp --dry-run copy_src.XXXXXX`
runTest $log 0 0 $null_src_path $dst_path
# if the copy failed
if [ 0 -ne $? ]; then
	# if the destination file does not exist
	if [ ! -f $dst_path ]; then
		printTestPass
		pass_count=`$EXPR $pass_count + 1`
	fi
fi
$RM -f $dst_path


#
# Test 4
# Full file copy from a source file that is not readable
# to a destination file that does not exist
# Expect to fail
#
test_count=`$EXPR $test_count + 1`
printTestStart $test_count

cmd="$CHMOD u-r $src_path"
$cmd
expectSuccess "$cmd" $?
# if chmod succeeded
if [ 0 -eq $? ]; then
	runTest $log 0 0 $null_src_path $dst_path
	# if the copy failed
	if [ 0 -ne $? ]; then
		# if the destination file does not exist
		if [ ! -f $dst_path ]; then
			printTestPass
			pass_count=`$EXPR $pass_count + 1`
		fi
	fi
fi
$CHMOD u+r $src_path
$RM -f $dst_path


#
# Summary
#
$ECHO ""
$ECHO "Test Summary"
$ECHO $SEPARATOR
$ECHO "Passed: $pass_count / $test_count"
$ECHO "   Log: $log"
