/*
 * A utility to invoke the copyfile (CF) system call.
 *
 *
 * Copyright 2010 NetApp, Inc., 
 * contribution by James Lentini and Anshul Madan 
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied 
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License v2 for more details.
 */

#include <errno.h>
#include <fcntl.h> /* for AT_FDCWD */
#include <libgen.h>
#include <limits.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/syscall.h>
#include <sys/types.h>

#ifdef __i386__
#define __NR_copyfileat		338
#else
#define __NR_copyfileat		300
#endif

struct cf_params {
	bool is_partial;
	bool is_quiet;
	size_t count;
	int flags;
	long long src_offset;
	long long dst_offset;
	char *src_path;
	char *dst_path;
};

/*
 * Print a usage summary
 */
static void cf_print_usage(char *argv0, char *error)
{
	char *dup, *prog;

	dup = strdup(argv0);
	prog = basename(dup);

	fprintf(stderr, 
		"Error: %s \n\n"
		"Usage: %s [-f <flag>] [-p] [-c <count>] [-s <src offset>] [-d <dst offset>] [-q] <source path> <destination path> \n"
		"where: \n"
		"\t -p : partial file copy \n"
		"\t flag : the flag value \n"
		"\t count : the number of bytes to copy (ignored unless -p is specified) \n"
		"\t src offset : the source file offset in bytes (ignored unless -p is specified) \n"
		"\t dst offset : the destination file offset in bytes (ignored unless -p is specified) \n"
		"\t -q : quiet mode, suppress diagnostic messages \n",
		error,
		prog);

	free(dup);
}

/*
 * Parse a size_t parameter
 */
static bool cf_parse_size_t(char *arg, size_t *out_val)
{
	long int val;

	val = strtol(arg, NULL, 0);
	if (LONG_MIN == val || LONG_MAX == val) {
		return false;
	} else {
		*out_val = (size_t) val;
		return true;
	}
}

/*
 * Parse an int parameter
 */
static bool cf_parse_int(char *arg, int *out_val)
{
	long int val;

	val = strtol(arg, NULL, 0);
	if (LONG_MIN == val || LONG_MAX == val) {
		return false;
	} else {
		*out_val = (int) val;
		return true;
	}
}

/*
 * Parse a long long parameter
 */
static bool cf_parse_long_long(char *arg, long long *out_val)
{
	long long val;

	val = strtoll(arg, NULL, 0);
	if (LLONG_MIN == val || LLONG_MAX == val) {
		return false;
	} else {
		*out_val = val;
		return true;
	}
}

/*
 * Parse the command line parameters
 */
static bool cf_parse_params(int argc, char *argv[], struct cf_params *params)
{
	bool rc;
	int c;

	memset(params, 0, sizeof *params);

	opterr = 0; /* disable getopt error messages */
	while (-1 != (c = getopt(argc, argv, "f:c:s:d:pq"))) {
		switch (c) {
		case 'f': /* flag */
			rc = cf_parse_int(optarg, &params->flags);
			if (!rc) {
				cf_print_usage(argv[0],
					       "invalid flag");
				goto exit;
			}
			break;

		case 'c': /* count */
			rc = cf_parse_size_t(optarg, &params->count);
			if (!rc) {
				cf_print_usage(argv[0],
					       "invalid count");
				goto exit;
			}
			break;

		case 's': /* source offset */
			rc = cf_parse_long_long(optarg, &params->src_offset);
			if (!rc) {
				cf_print_usage(argv[0],
					       "invalid source offset");
				goto exit;
			}
			break;

		case 'd': /* destination offset */
			rc = cf_parse_long_long(optarg, &params->dst_offset);
			if (!rc) {
				cf_print_usage(argv[0],
					       "invalid destination offset");
				goto exit;
			}
			break;

		case 'p': /* partial file copy */
			params->is_partial = true;
			break;

		case 'q': /* quiet mode */
			params->is_quiet = true;
			break;

		default:
			cf_print_usage(argv[0], "unrecognized flag");
			rc = false;
			goto exit;
		}
	}

	/* if there are not exactly two remaining arguments */
	if (2 != (argc - optind)) {
		cf_print_usage(argv[0], "invalid number of arguments");
		rc = false;
		goto exit;
	}

	params->src_path = argv[optind++];
	params->dst_path = argv[optind++];

	rc = true;
exit:
	return rc;
}

void cf_message(struct cf_params *params, const char *format, ...)
{
	va_list args;

	if (!params->is_quiet) {
		va_start(args, format);
		vfprintf(stdout, format, args);
		va_end(args);
	}
}

int main(int argc, char *argv[])
{
	struct cf_params params;
	int rc;

	if (!cf_parse_params(argc, argv, &params)) {
		goto exit;
	}

	if (!params.is_partial) {
		loff_t lof = 0;

		rc = syscall(__NR_copyfileat,
			     AT_FDCWD,
			     params.src_path,
			     AT_FDCWD,
			     params.dst_path,
			     params.flags,
			     &lof);
		if (0 != rc) {
			fprintf(stderr, "copyfile(2) failed: %s (%d)\n",
				strerror(errno), errno);
		}

		cf_message(&params, "copied %lld bytes\n", (long long int) lof);
	} else { /* partial file copy */
		fprintf(stderr, "Error: partial file copy not implemented\n");
		rc = -1;
	}

exit:
	return rc;
}
